#include <string>

#include <gtest/gtest.h>

#include <userver/engine/run_standalone.hpp>
#include <userver/formats/json/serialize.hpp>
#include <userver/formats/json/value_builder.hpp>

USERVER_NAMESPACE_BEGIN

namespace {

std::string MakeStringOfDeepObject(std::size_t depth) {
  std::string str;
  str.reserve(depth * 6 + 1);
  for (std::size_t i = 0; i < depth; ++i) {
    str += R"({"a":)";
  }
  str += "1";
  for (std::size_t i = 0; i < depth; ++i) {
    str += "}";
  }
  return str;
}

std::string MakeStringOfDeepArray(std::size_t depth) {
  std::string str;
  str.reserve(2 * depth + 1);
  for (std::size_t i = 0; i < depth; ++i) {
    str += "[";
  }
  str += "1";
  for (std::size_t i = 0; i < depth; ++i) {
    str += "]";
  }
  return str;
}

template <typename Func>
void RunStackLimitedTest(Func&& fn) {
  constexpr std::size_t kWorkerThreads = 1;
  engine::TaskProcessorPoolsConfig config;
  config.coro_stack_size = 128 * 1024ULL;

  constexpr std::size_t kDepth = 16 * 1024;

  engine::RunStandalone(kWorkerThreads, config,
                        [fn = std::forward<Func>(fn)] { fn(kDepth); });
}

}  // namespace

TEST(FormatsJson, DeepObjectFromString) {
  RunStackLimitedTest([](std::size_t depth) {
    EXPECT_THROW(formats::json::FromString(MakeStringOfDeepObject(depth)),
                 formats::json::ParseException);
  });
}

TEST(FormatsJson, DeepArrayFromString) {
  RunStackLimitedTest([](std::size_t depth) {
    EXPECT_THROW(formats::json::FromString(MakeStringOfDeepArray(depth)),
                 formats::json::ParseException);
  });
}

TEST(FormatsJson, DeepObjectFromStringDuplicateKeys) {
  RunStackLimitedTest([](std::size_t depth) {
    auto json_str = MakeStringOfDeepObject(depth);
    // {a: {a: 1}} -> {a: {a: 1}
    json_str.pop_back();
    // {a: {a: 1} -> {a: {a: 1}, a: 1}
    json_str.append(R"(,"a":1})");

    EXPECT_THROW(formats::json::FromString(json_str),
                 formats::json::ParseException);
  });
}

TEST(FormatsJson, DeepObjectFromStringNonsense) {
  RunStackLimitedTest([](std::size_t depth) {
    auto json_str = MakeStringOfDeepObject(depth);
    // {a: {a: 1}} -> {a: {a: 1}
    json_str.pop_back();
    // {a: {a: 1} -> {a: {a: 1}, nonsense}
    json_str.append(R"(,"nonsense"})");

    EXPECT_THROW(formats::json::FromString(json_str),
                 formats::json::ParseException);
  });
}

USERVER_NAMESPACE_END
